"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.migrateSchema = exports.diffSchema = void 0;
const clc = require("colorette");
const sql_formatter_1 = require("sql-formatter");
const types_1 = require("./types");
const client_1 = require("./client");
const connect_1 = require("../gcp/cloudsql/connect");
const prompt_1 = require("../prompt");
const logger_1 = require("../logger");
const error_1 = require("../error");
const projectUtils_1 = require("../projectUtils");
const utils_1 = require("../utils");
const errors = require("./errors");
async function diffSchema(schema) {
    const { serviceName, instanceName, databaseId } = getIdentifiers(schema);
    await ensureServiceIsConnectedToCloudSql(serviceName, instanceName, databaseId);
    try {
        await (0, client_1.upsertSchema)(schema, true);
    }
    catch (err) {
        const invalidConnectors = errors.getInvalidConnectors(err);
        if (invalidConnectors.length) {
            displayInvalidConnectors(invalidConnectors);
        }
        const incompatible = errors.getIncompatibleSchemaError(err);
        if (incompatible) {
            displaySchemaChanges(incompatible);
            return incompatible.diffs;
        }
    }
    (0, utils_1.logLabeledSuccess)("dataconnect", `Database schema is up to date.`);
    return [];
}
exports.diffSchema = diffSchema;
async function migrateSchema(args) {
    const { options, schema, allowNonInteractiveMigration, validateOnly } = args;
    const { serviceName, instanceId, instanceName, databaseId } = getIdentifiers(schema);
    await ensureServiceIsConnectedToCloudSql(serviceName, instanceName, databaseId);
    try {
        await (0, client_1.upsertSchema)(schema, validateOnly);
        logger_1.logger.debug(`Database schema was up to date for ${instanceId}:${databaseId}`);
    }
    catch (err) {
        const incompatible = errors.getIncompatibleSchemaError(err);
        const invalidConnectors = errors.getInvalidConnectors(err);
        if (!incompatible && !invalidConnectors.length) {
            throw err;
        }
        const shouldDeleteInvalidConnectors = await promptForInvalidConnectorError(options, invalidConnectors, validateOnly);
        if (!shouldDeleteInvalidConnectors && invalidConnectors.length) {
            const cmd = suggestedCommand(serviceName, invalidConnectors);
            throw new error_1.FirebaseError(`Command aborted. Try deploying compatible connectors first with ${clc.bold(cmd)}`);
        }
        const migrationMode = incompatible
            ? await promptForSchemaMigration(options, databaseId, incompatible, allowNonInteractiveMigration)
            : "none";
        if (migrationMode === "none" && incompatible) {
            throw new error_1.FirebaseError("Command aborted.");
        }
        let diffs = [];
        if (incompatible) {
            diffs = await handleIncompatibleSchemaError({
                options,
                databaseId,
                instanceId,
                incompatibleSchemaError: incompatible,
                choice: migrationMode,
            });
        }
        if (invalidConnectors.length) {
            await deleteInvalidConnectors(invalidConnectors);
        }
        await (0, client_1.upsertSchema)(schema, validateOnly);
        return diffs;
    }
    return [];
}
exports.migrateSchema = migrateSchema;
function getIdentifiers(schema) {
    var _a, _b;
    const databaseId = (_a = schema.primaryDatasource.postgresql) === null || _a === void 0 ? void 0 : _a.database;
    if (!databaseId) {
        throw new error_1.FirebaseError("Schema is missing primaryDatasource.postgresql?.database, cannot migrate");
    }
    const instanceName = (_b = schema.primaryDatasource.postgresql) === null || _b === void 0 ? void 0 : _b.cloudSql.instance;
    if (!instanceName) {
        throw new error_1.FirebaseError("tried to migrate schema but instance name was not provided in dataconnect.yaml");
    }
    const instanceId = instanceName.split("/").pop();
    const serviceName = schema.name.replace(`/schemas/${types_1.SCHEMA_ID}`, "");
    return {
        databaseId,
        instanceId,
        instanceName,
        serviceName,
    };
}
function suggestedCommand(serviceName, invalidConnectorNames) {
    const serviceId = serviceName.split("/")[5];
    const connectorIds = invalidConnectorNames.map((i) => i.split("/")[7]);
    const onlys = connectorIds.map((c) => `dataconnect:${serviceId}:${c}`).join(",");
    return `firebase deploy --only ${onlys}`;
}
async function handleIncompatibleSchemaError(args) {
    const { incompatibleSchemaError, options, instanceId, databaseId, choice } = args;
    if (incompatibleSchemaError.destructive && choice === "safe") {
        throw new error_1.FirebaseError("This schema migration includes potentially destructive changes. If you'd like to execute it anyway, rerun this command with --force");
    }
    const projectId = (0, projectUtils_1.needProjectId)(options);
    const iamUser = await (0, connect_1.setupIAMUser)(instanceId, databaseId, options);
    const commandsToExecute = incompatibleSchemaError.diffs
        .filter((d) => {
        switch (choice) {
            case "all":
                return true;
            case "safe":
                return !d.destructive;
            case "none":
                return false;
        }
    })
        .map((d) => d.sql);
    if (commandsToExecute.length) {
        await (0, connect_1.execute)([
            `SET ROLE "${(0, connect_1.firebaseowner)(databaseId)}"`,
            ...commandsToExecute,
            `GRANT ALL PRIVILEGES ON ALL TABLES IN SCHEMA "public" TO PUBLIC`,
        ], {
            projectId,
            instanceId,
            databaseId,
            username: iamUser,
        });
        return incompatibleSchemaError.diffs;
    }
    return [];
}
async function promptForSchemaMigration(options, databaseName, err, allowNonInteractiveMigration) {
    displaySchemaChanges(err);
    if (!options.nonInteractive) {
        const choices = err.destructive
            ? [
                { name: "Execute all changes (including destructive changes)", value: "all" },
                { name: "Abort changes", value: "none" },
            ]
            : [
                { name: "Execute changes", value: "safe" },
                { name: "Abort changes", value: "none" },
            ];
        return await (0, prompt_1.promptOnce)({
            message: `Would you like to execute these changes against ${databaseName}?`,
            type: "list",
            choices,
        });
    }
    else if (!allowNonInteractiveMigration) {
        logger_1.logger.error("Your database schema is incompatible with your Data Connect schema. Run `firebase dataconnect:sql:migrate` to migrate your database schema");
        return "none";
    }
    else if (options.force) {
        return "all";
    }
    else if (!err.destructive) {
        return "safe";
    }
    else {
        logger_1.logger.error("This schema migration includes potentially destructive changes. If you'd like to execute it anyway, rerun this command with --force");
        return "none";
    }
}
async function promptForInvalidConnectorError(options, invalidConnectors, validateOnly) {
    if (!invalidConnectors.length) {
        return false;
    }
    displayInvalidConnectors(invalidConnectors);
    if (validateOnly) {
        return false;
    }
    else if (options.force ||
        (!options.nonInteractive &&
            (await (0, prompt_1.confirm)(Object.assign(Object.assign({}, options), { message: "Would you like to delete and recreate these connectors?" }))))) {
        return true;
    }
    return false;
}
async function deleteInvalidConnectors(invalidConnectors) {
    return Promise.all(invalidConnectors.map(client_1.deleteConnector));
}
function displayInvalidConnectors(invalidConnectors) {
    const connectorIds = invalidConnectors.map((i) => i.split("/").pop()).join(", ");
    (0, utils_1.logLabeledWarning)("dataconnect", `The schema you are deploying is incompatible with the following existing connectors: ${connectorIds}.`);
    (0, utils_1.logLabeledWarning)("dataconnect", `This is a ${clc.red("breaking")} change and will cause a brief downtime.`);
}
async function ensureServiceIsConnectedToCloudSql(serviceName, instanceId, databaseId) {
    let currentSchema;
    try {
        currentSchema = await (0, client_1.getSchema)(serviceName);
    }
    catch (err) {
        if (err.status === 404) {
            currentSchema = {
                name: `${serviceName}/schemas/${types_1.SCHEMA_ID}`,
                source: {
                    files: [],
                },
                primaryDatasource: {
                    postgresql: {
                        database: databaseId,
                        cloudSql: {
                            instance: instanceId,
                        },
                    },
                },
            };
        }
        else {
            throw err;
        }
    }
    if (!currentSchema.primaryDatasource.postgresql ||
        currentSchema.primaryDatasource.postgresql.schemaValidation === "STRICT") {
        return;
    }
    currentSchema.primaryDatasource.postgresql.schemaValidation = "STRICT";
    try {
        await (0, client_1.upsertSchema)(currentSchema, false);
    }
    catch (err) {
        if (err.status >= 500) {
            throw err;
        }
        logger_1.logger.debug(err);
    }
}
function displaySchemaChanges(error) {
    const message = "Your new schema is incompatible with the schema of your CloudSQL database. " +
        "The following SQL statements will migrate your database schema to match your new Data Connect schema.\n" +
        error.diffs.map(toString).join("\n");
    (0, utils_1.logLabeledWarning)("dataconnect", message);
}
function toString(diff) {
    return `\/** ${diff.destructive ? clc.red("Destructive: ") : ""}${diff.description}*\/\n${(0, sql_formatter_1.format)(diff.sql, { language: "postgresql" })}`;
}
